%{
/* Nenad Taskovic 5/93 e-mail: neyt@fon.bg.ac.yu www: http://fon.fon.bg.ac.yu/~neyt/ */

#include <stdio.h>
#include <string.h>

#define  CEO_BROJ       0  /* konstante za tip 'tip' koje oznacavaju ceo broj i */
#define  ZNAKOVNI_NIZ   1  /* znakovni niz */

#ifdef __TURBOC__
#  define strcasecmp stricmp  /* u Borland Turbo C-u funkcija za poredjenje stringova bez obzira na velika i mala slova */
#endif                      /* se zove 'stricmp' a na LINUX-u 'strcasecmp' */

typedef struct Element {  /* struktura elementa u listi elemenata */
   char ime[32];           /* ime elementa */
   int tip;                /* tip elementa */
   int duzina;             /* duzina elementa */
   struct Element *sled;   /* pokazivac na sledeci element u listi */
} Element;

typedef struct Lista {    /* struktura cvora koji objedinjuje ime liste i njene elemente u listu listi */
   char ime[32];          /* ime liste */
   Element *elementi;     /* pokazivac na listu elemenata */
   struct Lista *sled;    /* pokazivac na sledeci cvor liste */
} Lista;

void ispisi_liste (Lista *);  /* ispisuje sadrzaj lista */
void obrisi_liste (Lista *);  /* dealocira memoriju koju zauzimaju liste */
void *alociraj_listu (const char *, Element *);  /* alocira memoriju za listu elemenata */
void dodaj_listu (Lista *, Lista *);  /* dodaje listu elemenata */
void *alociraj_element (const char *, int, int); /* alocira memoriju za element */
void dodaj_element (Element *, Element *);  /* dodaje element u listu elemenata */

%}

%union {
   int integer;
   char id[32];
   void *pok;
}

%token <integer> INTEGER
%token <id> ID
%token INT
%token STRING
%type <integer> tip_elementa
%type <pok> element
%type <pok> lista
%type <pok> lista_komponenti
%type <pok> liste
%start pocetak

%%

pocetak     : liste                                {ispisi_liste((Lista *) $1); obrisi_liste((Lista *) $1); }
            ;

liste       : lista                                { $$ = $1; } /* postavlja glavu 'liste' na prvu podlistu 'lista' */
            | liste lista                          { dodaj_listu((Lista *) $1, (Lista *) $2); }
            ;

lista       : ID ':' '<' lista_komponenti '>'      { $$ = alociraj_listu($1, (Element *) $4); }
            ;

lista_komponenti  : element                        { $$ = $1; } /* postavlja glavu liste na prvi element */
                  | lista_komponenti ',' element   { dodaj_element((Element *) $1, (Element *) $3); }
                  ;

element     : ID ':' tip_elementa '(' INTEGER ')'  { $$ = alociraj_element($1, $3, $5); }
            ;

tip_elementa   : INT                               { $$ = CEO_BROJ; }
               | STRING                            { $$ = ZNAKOVNI_NIZ; }
               ;

%%

char *tipstr[2] = {"INT", "STRING"};

extern FILE *yyin, *yyout;  /* yyin - ulazna datoteka, yyout - izlazna datoteka preuzete iz modula za leksicku analizu */

void ispisi_liste (Lista *glv)
{
   Element *tek;

   while (glv) {  /* ciklus koji prolazi kroz listu listi */
      fprintf(yyout, "Lista '%s':\n<\n", glv->ime);
      tek = glv->elementi;
      while (tek) {  /* ciklus koji prolazi kroz listu elemenata */
         fprintf(yyout, "\t%s: %s(%d)\n", tek->ime, tipstr[tek->tip], tek->duzina);
         tek = tek->sled;
      }
      fprintf(yyout, ">\n\n");
      glv = glv->sled;
   }
}

void obrisi_liste (Lista *lglv)
{
   Lista *lpom;
   Element *eglv, *epom;

   while (lglv) {  /* ciklus koji prolazi kroz listu listi */
      eglv = lglv->elementi;
      while (eglv) {  /* ciklus koji prolazi kroz listu elemenata */
         epom = eglv;
         eglv = eglv->sled;
         free(epom);  /* brise element */
      }
      lpom = lglv;
      lglv = lglv->sled;
      free(lpom);  /* brise cvor liste koji pokazuje na upravo obrisanu podlistu elemenata */
   }
}

void *alociraj_listu (const char *ime , Element *el)
{
   Lista *li;

   li = (Lista *) malloc(sizeof(Lista));  /* alocira memoriju za cvor liste koji ce pokazivati na podlistu elemenata */
   strcpy(li->ime, ime);  /* upisuje u novo alociranu memoriju ime podliste */
   li->elementi = el;     /* upisuje vrednost pokazivaca koji pokazuje na podlistu (listu elemenata) */
   li->sled = (Lista *) NULL;  /* pokazivac na sledeci cvor liste postavlja na NULL */

   return (void *) li;  /* vraca pokazivac na novo alocirani cvor liste */
}

void dodaj_listu (Lista *glv, Lista *cvor)
{
   int razliciti;

   while ( (glv->sled) && (razliciti = strcasecmp(glv->ime, cvor->ime)) )  /* proverava da li u listi postoji cvor */
      glv = glv->sled;    /* koji ima isto ime kao i cvor koji se ubacuje */

   if (razliciti)  /* ako je cvor koji se unosi jedinstven povezuje se u listu */
      glv->sled = cvor;
   else {
      Element *tek = cvor->elementi, *pom;
      
      while (tek) {  /* ako nije brise se cela podlista elemenata i cvor koji je trebalo ubaciti */
	 pom = tek;
	 tek = tek->sled;
	 free(pom);
      }
      free(cvor);
   }
}

void *alociraj_element (const char *ime, int tip, int duzina)
{
   Element *el;

   el = (Element *) malloc(sizeof(Element));  /* alocira memoriju za element liste */
   strcpy(el->ime, ime);  /* u alociranu memoriju kopira naziv elementa */
   el->tip = tip;         /* upisuje tip elementa */
   el->duzina = duzina;   /* upisuje duzinu elementa */
   el->sled = (Element *) NULL;  /* upisuje NULL u pokazivac na sledeci element */
   
   return (void *) el;  /* vraca pokazivac na formirani element */
}

void dodaj_element (Element *glv, Element *el)
{
   int razliciti;

   while ( (glv->sled) && (razliciti = strcasecmp(glv->ime, el->ime)) )  /* proverava da li postoji jos neki element sa */
      glv = glv->sled;   /* istim imenom u listi elemenata */
   if (razliciti)  /* ako ne postoji element se vezuje u listu */
      glv->sled = el;
   else
      free(el);   /* ako postoji brise se element koji je trebalo ubaciti u listu */
}

/* 
 * 8.  Napisati LEX i YACC speifikaije za sledecu gramatiku:
 *
 * pocetak     :  liste
 *             ;
 * liste       :  lista
 *             |  liste lista
 *             ;
 * lista       :  ID ':' '<' lista_komponenti '>'
 *             ;
 * lista_komponenti  :  element
 *                   |  lista_komponenti ',' element
 *                   ;
 * element     :  ID : tip
 *             ;
 * tip         :  INT '('duzina')'
 *             |  STRING '(' duzina ')'
 *             ;
 * duzina      :  INTEGER
 *             ;
 *
 *    Tip i vrednost tokena ID i INTEGER definisu se u LEX speifikaciji. ID
 * je identifikator koji se sastoji od niza znakova gde prvi znak obavezno
 * mora biti  slovo engeske azbuke, a ostali znaci u nizu su kombinacija
 * slova i cifara. INTEGER je ceo pozitian broj.
 * 
 *    U YACC speifikaciji, koriscenjem semantickih akcija treba napisati kod
 * za formiranje memorijske liste. Svaki cvor memorijske liste sadrzi tri
 * polja: prvo polje sadrzi ime elementa lista, drugo polje sadrzi tip
 * elementa, trece polje sadrzi duzinu, a cetvro polje je pokazivacko
 * polje. U listi se jedna promenljiva moze pojaviti samo jednom. Posle
 * formiranja jedne memorijske liste, treba prikazati: naziv liste i
 * elemente liste.
 */


int yyerror (char *s)
{
   extern int linija;  /* brojac linija preuzet iz modula za leksicku analizu */

   fprintf(stderr, "Greska: `%s'\nu liniji broj %d.\n", s, linija);
}

/*    Program je napisan tako da ga je moguce prevesti GNU C-om (pod Linux-om) i
 * Borland Turbo C-om pod DOS-om.
 *    Tekst koji treba da parsira program cita sa standardnog ulaza (stdin) a izlaz
 * daje na standardnom izlazu (stdout). Greske ispisuje na standardnom izlazu za
 * greske (stderr). Ovo se moze promeniti tako sto se kao parametri daju ulazna i
 * izlazna datoteka, sto od prilike izgleda ovako:
 * <ime_programa> [ <ulazna_datoteka> [ <izlazna_datoteka> ] ] <ENTER>
 *    Parser je napravljen tako da ne vodi racuna o velikim i malim slovima.
 * To je postignuto navodjenjem opcije '-i' pri pozivanju flex-a.
 *    Postoji mala razlika izmedju gramatike zadate u zadatku i iskoriscene gramatike.
 * Razlika je u tome sto je eliminisan neterminal `duzina'. I pored toga funkcionalnost
 * parsera je ostala ista.
 */

int main (int argc, char *argv[])
{
   if (argc >= 2) {  /* ako postoji dva ili vise argumenata (prvi je uvek naziv programa) */
      yyin = fopen(argv[1], "rt");  /* drugi se smatra nazivom ulazne datoteke */
      if (!yyin) {
         fprintf(stderr, "Ne mogu da otvorim datoteku sa imenom '%s'.\n", argv[1]);
         return (1);
      }
   }

   if (argc >= 3) {  /* ako postoji tri ili vise argumenata */
      yyout = fopen(argv[2], "wt"); /* treci se smatra nazivom izlazne datoteke */
      if (!yyout) {
	      fprintf(stderr, "Ne mogu da kreiram datoteku sa imenom '%s'.\n", argv[2]);
	      return (2);
      }
   }

   return yyparse(); /* poziv parsera koji kao rezultat vraca kod greske, sto je u ovom slucaju */
                     /* vrednost koju program vraca operativnom sistemu kao indikator uspesnosti */
                     /* izvrsenja programa. (0 - uspesno, !0 - neuspesno). */
}
